# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Numpy functions not available in the minimal required version.

Current minimum Numpy version: 1.13.3
"""

from __future__ import absolute_import, division, print_function

__all__ = ()

if __name__ == '__main__':
    from odl.util.testutils import run_doctests
    run_doctests()
